﻿using System;
using System.Data;
using System.Configuration;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;

using System.Web.Configuration;
using System.Data.SqlClient;

namespace com.RPTI.DAL
{
    [System.ComponentModel.DataObject]
    public class MultiRowDisConn_DAL
    {
        private string strConnString = System.Configuration.ConfigurationManager.ConnectionStrings["ConnString_SqlClient"].ToString();

        public MultiRowDisConn_DAL()
        {
        }

        #region ****** GridView 在 DAL 層，執行 SELECT 時的函數 (只有頁面第一次進入時、存檔成功後會使用到) ******

        [System.ComponentModel.DataObjectMethodAttribute(System.ComponentModel.DataObjectMethodType.Select, true)]  // true 表示為預設值
        public DataTable SelectDataSetBindGridView()
        {
            SqlConnection odConn = null;
            SqlCommand odCmd = null;
            SqlDataAdapter odAdapter = null;
            DataTable dt = null;
            string strSql = "SELECT * from table1 order by id desc";    // id為主鍵(Identity)

            try
            {
                odConn = new SqlConnection(strConnString);
                odConn.Open();

                if (odConn.State == ConnectionState.Open)
                {
                    dt = new DataTable();
                    odCmd = new SqlCommand(strSql, odConn);
                    //odCmd.Parameters.Add("@DATA_ID", SqlDbType.Float, intColumnLength).Value = strTheNewestDATA_ID;   //若SQL有參數條件

                    odAdapter = new SqlDataAdapter(odCmd);
                    odAdapter.Fill(dt);
                }
            }
            catch (SqlException ex)
            {
                throw new Exception("發生資料庫存取錯誤: " + ex.Message);
            }
            catch (Exception ex)
            {
                throw new Exception("發生資料庫存取以外的錯誤: " + ex.Message);
            }
            finally
            {
                if (odCmd != null)
                    odCmd.Dispose();
                if (odAdapter != null)
                    odAdapter.Dispose();
                if (dt != null)
                    dt.Dispose();
                if (odConn.State == ConnectionState.Open)
                {
                    odConn.Close();
                }
                odConn.Dispose();
            }

            return dt;

        } // end of func

        #endregion


        #region ****** 對多筆離線編輯的資料，「依序」做「修改、刪除、新增」 ******

        // 若寫入資料庫全部成功，會回傳 1；若發生「多人並行衝突」，會回傳 -0
        public int UpdateGridView(DataTable dtUpdate, DataTable dtInsert)
        {
            int intUpdatedRowCount = 0;         //「修改、刪除、新增」成功後，回傳受影響的資料列數。若發生「多人並行存取」修改失敗，值會保持 0；若寫入資料庫發生錯誤，值為 -1
            SqlConnection odConn = null;
            SqlCommand odCmdUpdate = null;              // step 1
            SqlCommand odCmdDelete = null;              // step 2
            SqlCommand odCmdUpdateMODIFY_DT = null;     // step 3
            SqlCommand odCmdInsert = null;              // step 4
            SqlTransaction odt = null;          // 資料庫交易
            string strSqlUpdate = string.Empty;
            string strSqlDelete = string.Empty;
            string strSqlUpdateMODIFY_DT = string.Empty;
            string strSqlInsert = string.Empty;
            int intDtUpdateRowCount = dtUpdate.Rows.Count;  // 欲修改的 DataTable 中的筆數，用來和「實際修改成功的筆數」比對之用。若兩者不符合，即為發生「並行存取衝突」
            int intDtDeleteRowCount = 0;    // 欲刪除的 DataTable 中的筆數(亦即 GridView1 勾選的筆數)，用來和「實際刪除成功的筆數」比對之用。若兩者不符合，即為發生「並行存取衝突」
            int intInsertRowCount = 0;  // 實際要新增的筆數(使用者在新增列，沒有勾選的總筆數)

            try
            {
                odConn = new SqlConnection(strConnString);
                odConn.Open();
                if (odConn.State == ConnectionState.Open)
                {
                    odt = odConn.BeginTransaction(IsolationLevel.ReadUncommitted);  // 設定資料庫的交易隔離等級，為完全不會造成任何資料列被鎖定

                    #region ****** step 1: 修改: 在 GridView1 上半部區塊裡，不論前面是否有勾選，將每一筆資料都存入資料庫 (並做 Concurrent 多人並行存取衝突偵測) ******

                    // 要等到 UPDATE、DELETE 都成功後，再將每筆的 MODIFY_DT 更新成最新的時間 (step 3)，
                    // 否則若 UPDATE 的同時就更新 MODIFY_DT，會造成後續 DELETE 的 concurrent 比對永遠不符合。

                    // 若資料表為全空、一筆資料都沒有時 (dtUpdate.Rows.Count = 0)，不會執行到這一段
                    if (dtUpdate.Rows.Count > 0)
                    {
                        odCmdUpdate = new SqlCommand();
                        odCmdUpdate.Connection = odConn;

                        for (int i = 0; i < dtUpdate.Rows.Count; i++)
                        {
                            // 因批次更新時，不同句子的 Parameter 參數名稱(@xxx)不能重複，因此手動在每個參數名稱後面，加上一個變動的 i 值，以組合出多句並存時，也不會重複的參數名稱
                            strSqlUpdate += "UPDATE table1 SET float1=@float1" + i + ", date1=@date1" + i + ", varchar1=@varchar1" + i + ", varchar2=@varchar2" + i +
                                " WHERE id=@id" + i + " AND convert(char(30), MODIFY_DT, 20)=@MODIFY_DT" + i + " ;";     // 最後用一個「分號(;)」串連多句 sql 語句，做批次更新

                            // 別支程式套用時，必須改以下五行的 Rows[i][n] 的「行」索引編號。可參考前端 .aspx 的 GridView_Update 的「欄位」位置 (從 0 開始)
                            odCmdUpdate.Parameters.Add("@float1" + i, SqlDbType.Float).Value = dtUpdate.Rows[i][1].ToString();
                            odCmdUpdate.Parameters.Add("@date1" + i, SqlDbType.DateTime).Value = dtUpdate.Rows[i][2].ToString();
                            odCmdUpdate.Parameters.Add("@varchar1" + i, SqlDbType.NVarChar, 20).Value = dtUpdate.Rows[i][3].ToString();
                            odCmdUpdate.Parameters.Add("@varchar2" + i, SqlDbType.NVarChar, 20).Value = dtUpdate.Rows[i][4].ToString();
                            odCmdUpdate.Parameters.Add("@id" + i, SqlDbType.Int).Value = dtUpdate.Rows[i][0].ToString();            // Primary Key
                            odCmdUpdate.Parameters.Add("@MODIFY_DT" + i, SqlDbType.DateTime).Value = dtUpdate.Rows[i]["MODIFY_DT"].ToString();  // 做 Concurrent 比對的欄位
                        }

                        odCmdUpdate.CommandText = strSqlUpdate;
                        odCmdUpdate.Transaction = odt;

                        intUpdatedRowCount = odCmdUpdate.ExecuteNonQuery();

                        // 如果「修改成功筆數」和「DataTable 中欲修改的筆數」不合，即表示發生 Concurrent (此時不會引發 Exception)
                        if (intUpdatedRowCount != intDtUpdateRowCount)
                        {
                            odt.Rollback();     // 回復「交易」
                            return 0;           // 註記發生「多人並行存取衝突」
                        }
                    }
                    //else
                    //{
                    //    return -2;      // 若為全空的資料表，一筆資料都還沒有，使用者就按下「存檔」按鈕
                    //}

                    #endregion ****** step 1: 修改: 在 GridView1 上半部區塊裡，不論前面是否有勾選，將每一筆資料都存入資料庫 (並做 Concurrent 多人並行存取衝突偵測) ******


                    #region ****** step 2: 刪除: 在 GridView1 上半部區塊裡，將前面有勾選的資料列，從資料庫裡刪除 (並做 Concurrent 多人並行存取衝突偵測) ******

                    // 若資料表為全空、一筆資料都沒有時 (dtUpdate.Rows.Count = 0)，不會執行到這一段
                    if (dtUpdate.Rows.Count > 0)
                    {
                        odCmdDelete = new SqlCommand();
                        odCmdDelete.Connection = odConn;

                        for (int i = 0; i < dtUpdate.Rows.Count; i++)
                        {
                            // 若這一筆，使用者有勾選要刪除
                            if (dtUpdate.Rows[i]["chk"].ToString() == "True")
                            {
                                // 最後用一個「分號(;)」串連多句 sql 語句，做批次刪除
                                strSqlDelete += "DELETE FROM table1 WHERE id=@id" + i + " AND convert(char(30), MODIFY_DT, 20) = @MODIFY_DT" + i + " ;";

                                // 別支程式套用時，必須改以下這一行的 Rows[i][n] 的「行」索引編號。可參考前端 .aspx 的 GridView_Update 的「欄位」位置 (從 0 開始)
                                odCmdDelete.Parameters.Add("@id" + i, SqlDbType.Int).Value = dtUpdate.Rows[i][0].ToString();        // Primary Key、做 Concurrent 比對的欄位
                                odCmdDelete.Parameters.Add("@MODIFY_DT" + i, SqlDbType.DateTime).Value = dtUpdate.Rows[i]["MODIFY_DT"].ToString();  // 做 Concurrent 比對的欄位

                                intDtDeleteRowCount++;    // 要刪除的筆數 + 1
                            }
                        }

                        // 必須要使用者至少有勾選一筆以上，才會去資料庫刪除
                        if (intDtDeleteRowCount > 0)
                        {
                            odCmdDelete.CommandText = strSqlDelete;
                            odCmdDelete.Transaction = odt;

                            intUpdatedRowCount = odCmdDelete.ExecuteNonQuery();

                            // 如果「刪除成功筆數」和「DataTable 中欲刪除的筆數」不合，即表示發生 Concurrent (此時不會引發 Exception)
                            if (intUpdatedRowCount != intDtDeleteRowCount)
                            {
                                odt.Rollback();     // 回復「交易」
                                return 0;           // 註記發生「多人並行存取衝突」
                            }
                        }
                    }

                    #endregion ****** step 2: 刪除: 在 GridView1 上半部區塊裡，將前面有勾選的資料列，從資料庫裡刪除 (並做 Concurrent 多人並行存取衝突偵測) ******


                    #region ****** step 3: 若前述「修改、刪除」都已成功，就將 GridView1 上半部區塊裡，每筆的 MODIFY_DT 更新成最新的時間 ******

                    // 若資料表為全空、一筆資料都沒有時 (dtUpdate.Rows.Count = 0)，不會執行到這一段
                    if (dtUpdate.Rows.Count > 0)
                    {
                        odCmdUpdateMODIFY_DT = new SqlCommand();
                        odCmdUpdateMODIFY_DT.Connection = odConn;

                        for (int i = 0; i < dtUpdate.Rows.Count; i++)   // GridView1 上半部區塊裡的筆數
                        {
                            strSqlUpdateMODIFY_DT += "UPDATE table1 SET MODIFY_DT=getdate() WHERE id=@id" + i + " ;";

                            // 別支程式套用時，必須改以下這一行的 Rows[i][n] 的「行」索引編號。可參考前端 .aspx 的 GridView_Update 的「欄位」位置 (從 0 開始)
                            odCmdUpdateMODIFY_DT.Parameters.Add("@id" + i, SqlDbType.Int).Value = dtUpdate.Rows[i][0].ToString();   // Primary Key
                        }

                        odCmdUpdateMODIFY_DT.CommandText = strSqlUpdateMODIFY_DT;
                        odCmdUpdateMODIFY_DT.Transaction = odt;

                        odCmdUpdateMODIFY_DT.ExecuteNonQuery();
                    }

                    #endregion ****** step 3: 若前述「修改、刪除」都已成功，就將 GridView1 上半部區塊裡，每筆的 MODIFY_DT 更新成最新的時間 ******


                    #region ****** step 4: 新增: 在 GridView1 下半部區塊裡，將前面有勾選的資料列，存入資料庫；若沒勾選則不處理 ******

                    // 若資料表為全空、一筆資料都沒有時 (dtUpdate.Rows.Count = 0)，仍會執行到這一段

                    if (dtInsert.Rows.Count > 0)    // 若使用者有新增至少一筆
                    {
                        odCmdInsert = new SqlCommand();
                        odCmdInsert.Connection = odConn;

                        for (int i = 0; i < dtInsert.Rows.Count; i++)
                        {
                            // 若這一筆，使用者沒有勾選要刪除 (預設值，亦即要新增這一筆)
                            if (dtInsert.Rows[i]["chk"].ToString() == "False")
                            {
                                // 最後用一個「分號(;)」串連多句 sql 語句，做批次新增
                                strSqlInsert += "INSERT INTO table1 (float1, date1, varchar1, varchar2, MODIFY_DT) " +
                                    "VALUES (@float1" + i + ", @date1" + i + ", @varchar1" + i + ", @varchar2" + i + ", getdate()) ;";

                                // 別支程式套用時，必須改以下四行的 Rows[i][n] 的「行」索引編號。可參考前端 .aspx 的 GridView_Insert 的「欄位」位置 (從 0 開始)
                                odCmdInsert.Parameters.Add("@float1" + i, SqlDbType.Float).Value = dtInsert.Rows[i][1].ToString();
                                odCmdInsert.Parameters.Add("@date1" + i, SqlDbType.DateTime).Value = dtInsert.Rows[i][2].ToString();
                                odCmdInsert.Parameters.Add("@varchar1" + i, SqlDbType.NVarChar, 20).Value = dtInsert.Rows[i][3].ToString();
                                odCmdInsert.Parameters.Add("@varchar2" + i, SqlDbType.NVarChar, 20).Value = dtInsert.Rows[i][4].ToString();

                                intInsertRowCount++;
                            }
                        }

                        odCmdInsert.CommandText = strSqlInsert;
                        odCmdInsert.Transaction = odt;

                        if (intInsertRowCount > 0)
                        {
                            odCmdInsert.ExecuteNonQuery();
                        }

                        // 若資料新增成功，且這是在資料表原本為全空、一筆資料都沒有時 (dtUpdate.Rows.Count = 0) 的情況下新增成功的
                        if (dtUpdate.Rows.Count == 0)
                        {
                            intUpdatedRowCount = 1;     // 註記有新增成功
                        }
                    }

                    #endregion ****** step 4: 新增: 在 GridView1 下半部區塊裡，將前面有勾選的資料列，存入資料庫；若沒勾選則不處理 ******


                    odt.Commit();       // 確認「交易」
                }
            }
            catch (SqlException ex)
            {
                odt.Rollback();     // 回復「交易」
                throw new Exception("發生資料庫存取錯誤: " + ex.Message);
            }
            catch (Exception ex)
            {
                odt.Rollback();     // 回復「交易」
                throw new Exception("發生資料庫存取以外的錯誤: " + ex.Message);
            }
            finally
            {
                if (dtUpdate != null)
                {
                    dtUpdate.Dispose();
                }
                if (dtInsert != null)
                {
                    dtInsert.Dispose();
                }

                if (odCmdUpdate != null)
                {
                    odCmdUpdate.Dispose();
                }
                if (odCmdDelete != null)
                {
                    odCmdDelete.Dispose();
                }
                if (odCmdUpdateMODIFY_DT != null)
                {
                    odCmdUpdateMODIFY_DT.Dispose();
                }
                if (odCmdInsert != null)
                {
                    odCmdInsert.Dispose();
                }
                
                if (odt != null)
                {
                    odt.Dispose();
                    odt = null;
                }
                if (odConn.State == ConnectionState.Open)
                {
                    odConn.Close();
                }

                odConn.Dispose();
            }

            return intUpdatedRowCount;

        } // end of function

        #endregion ******  ******


    } // end of class
}
    
